// Copyright (c) 2024 Orange. All rights reserved.
// This software is distributed under the BSD 3-Clause-clear License, the text of which is available
// at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

#pragma once

#include "KWDRReference.h"
#include "KWDRRandom.h"
#include "KWDRStandard.h"
#include "KWDRMath.h"
#include "KWDRCompare.h"
#include "KWDRLogical.h"
#include "KWDRDateTime.h"
#include "KWDRString.h"
#include "KWDRText.h"
#include "KWDRTextList.h"
#include "KWDRVector.h"
#include "KWDRHashMap.h"
#include "KWDRMultiTable.h"
#include "KWDRTextualAnalysis.h"

// Enregistrement de toutes les regles de derivation
void KWDRRegisterAllRules();
