// Copyright (c) 2024 Orange. All rights reserved.
// This software is distributed under the BSD 3-Clause-clear License, the text of which is available
// at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

#pragma once

//DDD #include "PLUseMPI.h"

#include "UserInterface.h"
#include "KWContinuous.h"
#include "KWSymbol.h"
#include "KWType.h"
#include "KWClass.h"
#include "KWClassDomain.h"
#include "KWDerivationRule.h"
#include "KWObject.h"
#include "KWStat.h"
#include "KWFrequencyVector.h"
#include "KWAttributeStats.h"
#include "FileService.h"
#include "KWDRAll.h"
#include "KWDiscretizerTest.h"
#include "KWDiscretizerMODL.h"
#include "KWGrouperMODL.h"
#include "KWGrouperTest.h"
#include "KWQuantileBuilder.h"
#include "KWStatisticalEvaluation.h"
#include "KWPredictorSpecView.h"
#include "KWPredictorSpecArrayView.h"
#include "KWDataGrid.h"
#include "StreamTest.h"
#include "KWDensityEstimationTest.h"
#include "KWDataGridTest.h"
#include "KWTextParser.h"
#include "KWSNBStudy.h"
#include "Timer.h"
#include "KWDataGridStats.h"
#include "KWProbabilityTable.h"
#include "KWObjectKey.h"
#include "KWSTDatabaseTextFile.h"
#include "KDMultinomialSamplingStudy.h"
#include "KDMultinomialSampleGenerator.h"
#include "KWDate.h"
#include "KWTime.h"
#include "KWTimestamp.h"
#include "KWTimestampTZ.h"
#include "KWTypeAutomaticRecognition.h"
#include "FileCache.h"
#include "InputBufferedFile.h"
#include "OutputBufferedFile.h"
#include "KWMetaData.h"
#include "Divers.h"
#include "CreateLargeFiles.h"
#include "SystemDivers.h"
#include "KWLearningProject.h"
#include "KWHierarchicalMultinomialStudy.h"
#include "KDConstructedRule.h"

#include "PLParallelTask.h"
#include "RMResourceManager.h"

#include "KWArtificialDataset.h"
#include "KWKeySizeEvaluatorTask.h"
#include "KWSortBuckets.h"
#include "KWFileKeyExtractorTask.h"
#include "KWKeySizeEvaluatorTask.h"
#include "KWKeySampleExtractorTask.h"
#include "KWSortedChunkBuilderTask.h"
#include "KWChunkSorterTask.h"
#include "PLSTDatabaseTextFile.h"
#include "KWTestDatabaseTransfer.h"
#include "KWKeyPositionSampleExtractorTask.h"
#include "KWKeyPositionFinderTask.h"
#include "KWLoadIndex.h"
#include "KWTupleTable.h"
#include "KWTupleTableLoader.h"
#include "KWDataTableSliceSet.h"
#include "JSONFile.h"
#include "JSONTokenizer.h"
#include "KWCDUniqueString.h"

#include "KWNewType.h"
#include "SizeofStudy.h"
#include "SparseStudy.h"
#include "KWValueBlock.h"
#include "KWDRTextualAnalysis.h"
#include "SNBIndexVector.h"
#include "KWTextService.h"
#include "KWTextTokenizer.h"
#include "KWResultFilePathBuilder.h"

#include "HanoiTower.h"
