// Copyright (c) 2024 Orange. All rights reserved.
// This software is distributed under the BSD 3-Clause-clear License, the text of which is available
// at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

// Fichier genere automatiquement pour l'affichage des ressources
// Ne pas modifier

#ifdef _WIN32
//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by MODL_Coclustering.rc

#include "../KWUtils/KWKhiopsVersion.h"

// Valeurs par defaut suivantes des nouveaux objets
//
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE 101
#define _APS_NEXT_COMMAND_VALUE 40001
#define _APS_NEXT_CONTROL_VALUE 1001
#define _APS_NEXT_SYMED_VALUE 101
#endif
#endif
#endif
