// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "winres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Anglais (États-Unis) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH,
    SUBLANG_ENGLISH_US

#ifdef APSTUDIO_INVOKED
    /////////////////////////////////////////////////////////////////////////////
    //
    // TEXTINCLUDE
    //

    1 TEXTINCLUDE BEGIN "resource.h\0" END

    2 TEXTINCLUDE BEGIN "#include "
			"winres.h"
			"\r\n"
			"\0" END

    3 TEXTINCLUDE BEGIN "\r\n"
			"\0" END

#endif // APSTUDIO_INVOKED

#endif // Anglais (États-Unis) resources
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Francais (France) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_FRA)
			    LANGUAGE LANG_FRENCH,
    SUBLANG_FRENCH

	/////////////////////////////////////////////////////////////////////////////
	//
	// Version
	//

	VS_VERSION_INFO VERSIONINFO FILEVERSION 0,
    0, 0, 0 PRODUCTVERSION 0, 0, 0,
    0 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
    FILEFLAGS 0x1L
#else
    FILEFLAGS 0x0L
#endif
    FILEOS 0x40004L FILETYPE 0x1L FILESUBTYPE 0x0L BEGIN BLOCK "StringFileInfo" BEGIN BLOCK "040c04b0" BEGIN VALUE
							       "CompanyName",
    "Orange" VALUE "FileDescription", "Khiops Native Interface Library" VALUE "FileVersion",
    KHIOPS_VERSION VALUE "InternalName", "KhiopsNativeInterface.dll" VALUE "LegalCopyright",
    KHIOPS_COPYRIGHT_LABEL VALUE "OriginalFilename", "KhiopsNativeInterface.dll" VALUE "ProductName",
    "Khiops Native Interface" VALUE "ProductVersion",
    KHIOPS_VERSION END END BLOCK "VarFileInfo" BEGIN VALUE "Translation", 0x40c, 1200 END END

/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

#endif // Francais (France) resources
/////////////////////////////////////////////////////////////////////////////

#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//

/////////////////////////////////////////////////////////////////////////////
#endif // not APSTUDIO_INVOKED