// Copyright (c) 2024 Orange. All rights reserved.
// This software is distributed under the BSD 3-Clause-clear License, the text of which is available
// at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

#include "DTGlobalTag.h"

const ALString DTGlobalTag::SCHRODER_LABEL = "SCHRODER";
const ALString DTGlobalTag::REC_LABEL = "REC";
const ALString DTGlobalTag::REC_RISSANEN_LABEL = "REC_RISSANEN";
const ALString DTGlobalTag::REC_RISSANEN_V1_LABEL = "REC_RISSANEN_V1";
const ALString DTGlobalTag::REC_RISSANEN_V2_LABEL = "REC_RISSANEN_V2";
const ALString DTGlobalTag::REC_N2_LABEL = "REC_N2";
const ALString DTGlobalTag::REC_BIN_LABEL = "REC_BIN";
const ALString DTGlobalTag::REC_NEW_BIN_LABEL = "REC_NEW_BIN";
const ALString DTGlobalTag::REC_TRI_LABEL = "REC_TRI";
const ALString DTGlobalTag::CATALAN_LABEL = "CATALAN";
const ALString DTGlobalTag::IDENTICAL_LABEL = "Identical";
const ALString DTGlobalTag::BEST_TREE_COST_LABEL = "Best tree cost";
const ALString DTGlobalTag::RANDOM_UNIFORM_LABEL = "Random uniform";
const ALString DTGlobalTag::RANDOM_LEVEL_LABEL = "Random level";
const ALString DTGlobalTag::MODL_LABEL = "MODL";
const ALString DTGlobalTag::MODL_EQUAL_FREQUENCY_LABEL = "EqualFrequency";
const ALString DTGlobalTag::BASIC_GROUPING_LABEL = "BasicGrouping";
const ALString DTGlobalTag::PRE_PRUNING_LABEL = "Pre-pruning";
const ALString DTGlobalTag::POST_PRUNING_LABEL = "Post-pruning";
const ALString DTGlobalTag::NO_PRUNING_LABEL = "No pruning";
const ALString DTGlobalTag::UNIFORM_SAMPLING_WITH_REPLACEMENT_LABEL = "Uniform sampling with replacement";
const ALString DTGlobalTag::RANK_WITH_REPLACEMENT_LABEL = "Rank sampling with replacement";
const ALString DTGlobalTag::LEVEL_SAMPLING_WITH_REPLACEMENT_LABEL = "Level sampling with replacement";
const ALString DTGlobalTag::NODE_UNIFORM_SAMPLING_WITH_REPLACEMENT_LABEL = "Node Uniform sampling with replacement";
const ALString DTGlobalTag::NODE_LEVEL_SAMPLING_WITH_REPLACEMENT_LABEL = "Node Level sampling with replacement";
const ALString DTGlobalTag::RF_OPTIMIZATION_RANDOM_VARIABLES = "Multi-start with random variables number";
const ALString DTGlobalTag::RF_OPTIMIZATION_ALL_TREES = "All trees";
const ALString DTGlobalTag::RF_OPTIMIZATION_BEST_TREE = "Best tree";
const ALString DTGlobalTag::RF_OPTIMIZATION_MULTISTART = "Multi-start";
