// Copyright (c) 2024 Orange. All rights reserved.
// This software is distributed under the BSD 3-Clause-clear License, the text of which is available
// at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

#pragma once
#define KHIOPS_STR(s) #s

// Constantes definies a la fois dans les sources et dans les ressources Visual Studio.
// Les ressources permettent d'afficher des informations sur les executables
// dans le TaskManager de Windows (par exemple)

// Version de Khiops
#define GENUM_VERSION KHIOPS_STR(1.0)

// Copyright
#define GENUM_COPYRIGHT_LABEL KHIOPS_STR((c)2024 Orange.)
